


////////////////////////////////////////////////////////////////
//  FHT INIT
////////////////////////////////////////////////////////////////

void FHTinit(int nn)
{   
    int index, j, s, k;
    double ang = 0;
    double omega = 2 * PI / nn;
    int lg2 = int(log(double(nn)) / log(2.0) + .5);
    double knn = 4.0 / nn;
    
    for (int i = 0; i < nn; i++)
    {
        // WINDOWS
        //hTab[i] = float(sin(ang / 2) * knn);            // SIN
        //hTab[i] = float(pow(sin(ang / 2), 2) * knn);    // SIN ^ 2    (togliere pow)
        hTab[i] = float(pow(sin(ang / 2), 12) * knn);    // SIN ^ 12    (togliere pow)
        //hTab[i] = float((1 - cos(ang)) * knn);          // HANNING
        //hTab[i] = float((.54 - .46 * cos(ang)) * knn);  // HAMMING
        //hTab[i] = float((.42 - .5 * cos(ang)) * knn);   // BLACKMAN

        // SIN & COS
        senTab[i] = float(sin(ang));
        cosTab[i] = float(cos(ang));
        ang += omega;

        // PERMUTATION TABLE
        index = i;
        j = 0;
        for (k = 0; k < lg2; k ++)
        {
            s = index >> 1;
            j = j + j + index - s - s; 
            index = s;
        }
        permTab[i] = j;  
    }
  
    initSamples = nn;
}



////////////////////////////////////////////////////////////////
//  FHT
////////////////////////////////////////////////////////////////

void FHT(int nn)
{
    if (nn > samplesMax) return;

    if (nn != initSamples) FHTinit(nn);

    int i, j, le, le2, k, mk, trig_ind, trig_inc;
    float ua, ub, v;
    float* A = aTab;
    float* B = bTab;
    float* C;

    // Samples from circular buffer to aTab 

	// ======================================================================================
    //	FAST SWEEP & PULSE
	// ======================================================================================
    if ((analisistype == ANTYPE_FASTSW) || (analisistype == ANTYPE_PULSE))
    {
        double knn = 2.0 / nn;

		if (analisistype == ANTYPE_FASTSW) knn *= 11.67; // Correzione di livello per FAST SWEEP  POW15=22.5 / POW10=30.5
		if (analisistype == ANTYPE_PULSE) knn *= 2.23;	 // Correzione di livello per PULSE

        int start = int(SAMPLES_PER_SEC * start_sampling_ms / 1000);
		int nsamp = int(SAMPLES_PER_SEC * sampling_ms / 1000);
		if (nsamp > nn) nsamp = nn;

		// -------------------------------------------------- sampling_ms is not used for ANTYPE_FASTSW
		if (analisistype == ANTYPE_FASTSW)	 nsamp = nn;

        

		readpos = pulseCbuffPos;
		while (readpos < 0) readpos += IN_CBUF_LEN;

        j = readpos + start;                             
        for (i = 0; i < nsamp; i++)
        {
            aTab[permTab[i]] = pCbufIn[j] * float(knn); 
            if ((j += numCH) >= IN_CBUF_LEN) j -= IN_CBUF_LEN;        
        }


        


		/*
		// ============================================================== ZERO CORRECTION ( START )  
		float corrStart = float(0);
		//                             the 500 samples ( 11.3mS )
		//                             preceding the trigger position
		//                             are used to correct the zero position					
		j = readpos + start - 500;
		while (readpos < 0) readpos += IN_CBUF_LEN;
		//
		for (i = 0; i < 500; i++)
		{
			corrStart += pCbufIn[j] * float(knn); 
			if ((j += numCH) >= IN_CBUF_LEN) j -= IN_CBUF_LEN;        
		}
		corrStart = corrStart / float(500);
		// ------------------------------------------------------- correct START position
		for (i = 0; i < nsamp; i++) aTab[permTab[i]] -= corrStart;
		// ===============================================================================================
		*/

		///*
		// ============================================================== ZERO CORRECTION ( START to END )
        float corrStart = float(0);
		float corrEnd = float(0);
        // ----------------------------------------- test 1000 samples ( 11mS ) before the trigger position
		j = readpos + start - 1000;
		while (readpos < 0) readpos += IN_CBUF_LEN;
		//
        for (i = 0; i < 1000; i++)
        {
            corrStart += pCbufIn[j] * float(knn); 
            if ((j += numCH) >= IN_CBUF_LEN) j -= IN_CBUF_LEN;        
        }
        corrStart = corrStart / float(1000);

		// ------------------------------------------ test 1000 samples ( 11mS ) around the end position					
		j = readpos + start + nsamp - 500;
		while (readpos < 0) readpos += IN_CBUF_LEN;
		//
        for (i = 0; i < 1000; i++)
        {
            corrEnd += pCbufIn[j] * float(knn); 
            if ((j += numCH) >= IN_CBUF_LEN) j -= IN_CBUF_LEN;        
        }
        corrEnd = corrEnd / float(1000);

        // ------------------------------------------------------ correct START to END zero positions	
        for (i = 0; i < nsamp; i++) aTab[permTab[i]] -= corrStart + ((corrEnd - corrStart) * i) / nsamp;
		// ================================================================================================
		//*/



        // ============================================================== SMOOTING TO ZERO
		if (analisistype == ANTYPE_FASTSW)
		{
			// ---------------------------- NO SMOOTING WITH FAST SWEEP
			//for (i = 0; i < nsamp; i++)
			//{
			//	float corr = float(nsamp - i) / nsamp;
			//	aTab[permTab[i]] *= (pow(corr, float(0.5)));
			//}
		}
		else
		{
			for (i = 0; i < nsamp; i++)
			{
				float corr = float(nsamp - i) / nsamp;
				aTab[permTab[i]] *= (corr * corr);
			}
		}
    

        // ============================================================== ALL ZERO TO NN
        for (i = nsamp; i < nn; i++) aTab[permTab[i]] = float(0);
  


    }

	// ======================================================================================
    // SPECTRUM
	// ======================================================================================
    else 
    {
        j = cbufPos - nn * numCH;                                // STEREO SELECT ?
        if (j < 0) j += IN_CBUF_LEN;
		
        // ZERO CORRECTION EVALUATION
        int jj = j;
        float corr = float(0);
        for (i = 0; i < nn; i++)
        {
            corr = corr + pCbufIn[jj] * hTab[i]; 
            if ((jj += numCH) >= IN_CBUF_LEN) jj -= IN_CBUF_LEN;
        }

		// CORREZIONE SPERIMENTALE PER OTTENERE I -120dB sotto ai dieci Hertz
		corr = corr * float(1.1);
        
        // FROM CIRCULAR BUFFER TO aTab WITH CORRECTION AND WINDOW
        for (i = 0; i < nn; i++)
        {
            aTab[permTab[i]] = (pCbufIn[j] - corr) * hTab[i];  
            if ((j += numCH) >= IN_CBUF_LEN) j -= IN_CBUF_LEN;
        }

    }

    // ======================================================================================
    //  FHT 
	// ======================================================================================
    le = 1;
    trig_inc = nn >> 1;
    do
    {
        trig_ind = 0;

        j = 0;
        le2 = le * 2;

        do
        {
            ua = cosTab[trig_ind];
            ub = senTab[trig_ind];
            trig_ind += trig_inc;

            i = j;
            mk = le2 - i;

            do
            {
                k = le + i;

                v = ua * A[k] + ub * A[mk];
                
                B[k] = A[i] - v;
                B[i] = A[i] + v;

                mk += le2;
                i += le2;
            }
            while (i < nn);

            j++;
        }
        while (j < le);

        C = A;
        A = B;
        B = C;

        le *= 2;
        trig_inc = trig_inc >> 1;
    }
    while (le < nn);
                          
                           
	// ======================================================================================
    //   Results to responseTab 
	// ======================================================================================
    
    float kb = float((band - 1) / 2.0);
        
    int peakresponse;
    
    if ((wavetype == WAVETYPE_NOISE) || (analisistype == ANTYPE_FASTSW) || (analisistype == ANTYPE_PULSE)) peakresponse = 0;
    else peakresponse = 1;


    int index;
    double val;
    if (xlogtype) 
    {
        
        double kf = freqmin * nn / SAMPLES_PER_SEC;
        double kk = log(freqmax / freqmin) / dispx;
        int oldindex = int(kf);
        int bandindex = int(oldindex * band);

        if (peakresponse)
        {
            int n = 0;
            double max = -9;
            for (j = 0; j <= dispx; j++)
            {
                index = int(kf * exp(kk * j));
                if (band != 1) index += int(index * kb);

                if  (index >= bandindex || j == dispx)
                {
                    for (i = oldindex + 1; i <= index; i++)
                    {
                        ua = A[i];
                        ub = A[nn - i];
                        val = sqrt( ua * ua + ub * ub );
                        if (val > max) max = val;
                        n++;
                    }
                    oldindex = index;
                    bandindex = int(oldindex * band);
                }
                
                if (calibratedResponse) max *= calibrateTab[j];
                if (n > 0) responseTab[j] += (max - responseTab[j]) * responsespeed;
                else responseTab[j] = -9;
                max = -9;               
                n = 0;
            }
        }
        else
        {
            int n = 0;
            double sum = 0;
            for (j = 0; j <= dispx; j++)
            {
                index = int(kf * exp(kk * j));
                if (band != 1) index += int(index * kb);

                if  (index >= bandindex || j == dispx)
                {
                    for (i = oldindex + 1; i <= index; i++)
                    {
                        ua = A[i];
                        ub = A[nn - i];
                        val = sqrt( ua * ua + ub * ub );

						// ----------------------------------------------------- FastSweep correction
						if (analisistype == ANTYPE_FASTSW) val *= (pow(i, 0.4644) );  //POW15=0.4644
						// ----------------------------------------------------- PULSE correction
                        if (analisistype == ANTYPE_PULSE) val *= i;

                        sum += val;
                        n++;   
                    }
                    oldindex = index;
                    bandindex = int(oldindex * band);
                }
                if (calibratedResponse) sum *= calibrateTab[j];
                if (n > 0) responseTab[j] += (sum / n - responseTab[j]) * responsespeed;
                else responseTab[j] = -9;
                sum = 0;
                n = 0;
            }
        }

    }
    else
    {
        double kf = double(nn) / SAMPLES_PER_SEC;
        double kk = (freqmax - freqmin) / dispx;
        int oldindex = int(kf * freqmin);
        int bandindex = int(oldindex * band);
        int n = 0;
        double max = -9;
        if (peakresponse)
        {
            for (j = 0; j <= dispx; j++)
            {
                index = int(kf * (freqmin + kk * j));
                if (band != 1) index += int(index * kb);

                if  (index >= bandindex || j == dispx)
                {
                    for (i = oldindex; i < index; i++)
                    {
                        ua = A[i + 1];
                        ub = A[nn - (i + 1)];
                        val = sqrt( ua * ua + ub * ub );
                        if (val > max) max = val; 
                        n++;
                    }
                    oldindex = index;
                    bandindex = int(oldindex * band);
                }
                if (calibratedResponse) max *= calibrateTab[j];
                if (n > 0) responseTab[j] += (max - responseTab[j]) * responsespeed;
                else responseTab[j] = -9;
                max = -9;
                n = 0;
            }
        }
        else
        {
            int n = 0;
            double sum = 0;
            for (j = 0; j <= dispx; j++)
            {
                index = int(kf * (freqmin + kk * j));
                if (band != 1) index += int(index * kb);

                if  (index >= bandindex || j == dispx)
                { 
                    for (i = oldindex; i < index; i++)
                    {
                        ua = A[i + 1];
                        ub = A[nn - (i + 1)];
                        val = sqrt( ua * ua + ub * ub );

                        // ----------------------------------------------------- FastSweep correction
						if (analisistype == ANTYPE_FASTSW) val *= (pow(i, 0.395) );
						// ----------------------------------------------------- PULSE correction
                        if (analisistype == ANTYPE_PULSE) val *= i;
                        
						sum += val;
                        n++;
                    }
                    oldindex = index;
                    bandindex = int(oldindex * band);
                }
                if (calibratedResponse) sum *= calibrateTab[j];
                if (n > 0) responseTab[j] += (sum / n - responseTab[j]) * responsespeed;
                else responseTab[j] = -9;
                sum = 0;
                n = 0;
            }

        }
    }
}


void ResetResponseTab()
{
    for (int i = 0; i < RESP_TAB_LEN; i++) responseTab[i] = 0;
}